local L = AceLibrary("AceLocale-2.2"):new("GhostPulse")

local pairs = _G.pairs


EnemyCastWatcher = {

	New = function(this, env, listenerFunction)
		-- check arguments
		if not env or type(env) ~= "table" then
			error("Must supply an environment tracker to EnemyCastWatcher:New()")
		end
		
		if not listenerFunction or type(listenerFunction) ~= "function" then
			error("Must supply a listener function to EnemyCastWatcher:New()")
		end
			
		-- globals
		frame = CreateFrame("Frame")
		
		frame.nextCheckTime = 0
		
		frame.enabled = false		
		frame.enabledAlerts = { }
				
		frame.environment = env
		frame.OnDetect = listenerFunction
		
		-- functions
		frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		
		frame.GetAlertTypes = function()
			return {
				["enemyspellcasttarget"] = L["Enemy Spell Casts (Target)"],
				["enemyspellcastfocus"] = L["Enemy Spell Casts (Focus)"],
				["enemyspellcastboss"] = L["Enemy Spell Casts (Bosses)"],
			}
		end
		
		-- common configuration
		frame.SetOptions = function(this, options)
			this.settings = options
		end
		
		frame.CheckEnv = function(this, alertType)
			return this:CheckCombat(alertType) and this:CheckGrouping(alertType) and
				   ( this.settings[alertType].whileMounted == true or not IsMounted() )
		end
		
		frame.CheckCombat = function(this, alertType)
			local rc = true
			
			if this.settings and this.settings[alertType].inCombat then
				rc = this.environment:PlayerIsInCombat()
			end
			
			return rc
		end

		frame.CheckGrouping = function(this, alertType)
			local rc = true
			local curGroupType = this.environment:GetPlayerGroupType()
			
			
			if curGroupType == "solo" then
				rc = this.settings[alertType].solo
			elseif curGroupType == "party" then
				rc = this.settings[alertType].party
			elseif curGroupType == "raid" then
				rc = this.settings[alertType].raid
			end
			
			return rc
		end
		
		frame.CanShowThing = function(this, alertType, thing)
			local rc = true
			
			
			if thing and alertType and this.settings[alertType] then
				if not this.enabledAlerts[alertType] then
					rc = false
				else
					if this.settings[alertType].whitelist and table.isEmpty(this.settings[alertType].whitelist) == false then
						rc = this.settings[alertType].whitelist[thing]
					else
						if this.settings[alertType].blacklist then
							if this.settings[alertType].blacklist[thing] then
								rc = false
							end
						end
					end
				end
			end
		
			return rc
		end
		
		frame.SetAlertEnabled = function(this, alertType, value)
			this.enabledAlerts[alertType] = value
		end
		
		-- watcher-specific function
		frame:SetScript("OnEvent", function(this, event, ...)
			if this.enabled then
				if event == "COMBAT_LOG_EVENT_UNFILTERED" then
					if ( this.enabledAlerts["enemyspellcastboss"] and this:CheckEnv("enemyspellcastboss") ) or
					   ( this.enabledAlerts["enemyspellcasttarget"] and this:CheckEnv("enemyspellcasttarget") ) or
					   ( this.enabledAlerts["enemyspellcastfocus"] and this:CheckEnv("enemyspellcastfocus") ) then
						this:NotifyCast(...)
					end
				end
			end
		end)
		
		frame.NotifyCast = function(this, timeStamp, event, srcGuid, srcName, srcFlags, destGuid, destName, destFlags, spellID, spellName, flags)
			if event == "SPELL_CAST_START" then
				local suffix = ""
				
				
				if this.enabledAlerts["enemyspellcastboss"] and this:CheckEnv("enemyspellcastboss") then
					if this:IsBoss(srcName) then
						suffix = "boss"
					end
				end
				
				if suffix == "" then
					if srcGuid == UnitGUID("target") then
						suffix = "target"
					elseif srcGuid == UnitGUID("focus") then
						suffix = "focus"
					end
				end
								
				if suffix ~= "" then
					if not UnitIsFriend("player", suffix) then
						if this:CheckEnv("enemyspellcast"..suffix) then
							if this:CanShowThing("enemyspellcast"..suffix, spellName) then
								local name, rank, icon, cost, isFunnel, powerType, castTime = GetSpellInfo(spellID)
	
								
								-- alert
								this:Alert(suffix, srcName, spellName, icon, castTime)
							end
						end
					end
				end
			end
		end

		frame.Alert = function(this, unit, unitName, spellName, icon, castTime)
			if this.OnDetect then
				if type(this.OnDetect) == "function" then
					local info = {
						["Unit"] = unit,
						["Name"] = unitName,
						["Spell"] = spellName,
						["CastTime"] = string.format("%0.1f", castTime / 1000),
						["Type"] = "enemyspellcast"..unit,
					}
					
					this:OnDetect("enemyspellcast"..unit, spellName, icon, info)
				else
					error("EnemyCastWatcher.OnDetect must be a function")
				end
			end
		end
		
		frame.IsBoss = function(this, name)
			local rc = false
			local prefix = "raid"
			local limit = 40
			local foundUnit = ""
			
			
			if GetNumRaidMembers() == 0 then
				prefix = ""
				limit = 0
				
				if GetNumPartyMembers() > 0 then
					prefix = "party"
					limit = 4
				end
			end
			
			if prefix ~= "" then
				for i = 1, limit do
					local unit = prefix..i.."target"
					
					
					if UnitName(unit) == name then
						foundUnit = unit
						break
					end
				end
			end
			
			if foundUnit ~= "" then
				if UnitLevel(foundUnit) == -1 then
					rc = true
				end
			end
			
			return rc
		end
		
		return frame
	end,
	
}